/*
 * Contributor(s): Natasha Noy noy@smi.stanford.edu
 */
package edu.stanford.smi.fmatoowl;

import java.util.*;

import edu.stanford.smi.protege.model.*;
import edu.stanford.smi.protege.util.*;
import edu.stanford.smi.protegex.owl.jena.*;
import edu.stanford.smi.protegex.owl.model.*;
import edu.stanford.smi.protegex.owl.model.impl.*;

public class OWLDLComponent {
	private KnowledgeBase _sourceKB = null;

	private Vector _classesInQueue = new Vector();
	
	private JenaOWLModel _owlModel = null;
	private boolean _fromFull = false;
	private String _fromFullString = _fromFull ? "FULL" : "DL";		

	private static final int DOMAIN = 1;
	private static final int RANGE = 2;
	
	private static Slot _directTemplateSlot = null;
	private static Slot _rdfsLabelSlot = null;
	
//	private Collection _processedClasses = new ArrayList();
	
	public OWLDLComponent (JenaOWLModel owlModel, KnowledgeBase sourceKb, boolean fromFull) {
		_owlModel = owlModel;
		_sourceKB = sourceKb;
		_directTemplateSlot = _sourceKB.getSlot(Model.Slot.DIRECT_TEMPLATE_SLOTS);
//		if (Constants.RDFS_LABEL_PROPERTY != null)
//			_rdfsLabelSlot = _sourceKB.getSlot(Constants.RDFS_LABEL_PROPERTY);
		_fromFull = fromFull;
		_fromFullString = _fromFull ? "FULL" : "DL";	
	}
	
	public void createDLComponent (String [] roots) {
		
		Log.getLogger().info("" + new Date () + " **** DL COMPONENT FIRST PASS from " + _fromFullString + "******");
		firstPass (roots);

		// Protege OWL does not allow you to add domain and range statements in the importing ontology; hence block this for now
//		if (!_fromFull) {
			Log.getLogger().info("" + new Date () + " **** DOMAINS  from " + _fromFullString + "******");
			createTopLevelDomainsAndRanges (DOMAIN);
	
			Log.getLogger().info("" + new Date () + " **** RANGES ******");
			createTopLevelDomainsAndRanges (RANGE);
//		}

		Log.getLogger().info("" + new Date () + " **** DL COMPONENT SECOND PASS " + _fromFullString + "******");
		secondPass ();
	}
	
	
	public void createTopLevelDomainsAndRanges(int domainOrRange) {
		Collection slots = _sourceKB.getSlots();
		Iterator s = slots.iterator();
		while (s.hasNext()) {
			Slot next = (Slot)s.next();
			if (next.isSystem()) continue;
			
			OWLProperty nextProperty = FMAtoOWL.getOWLProperty(next);
			if (nextProperty == null) continue;
			
			Collection nextDomainsOrRanges = null;
			if (domainOrRange == DOMAIN) {
				//*** the following should work, but there is a Protege bug, and for many slots it returns an incomplete list
				// (e.g. regional part doesn't have Anatomical structure in its domain)
				//nextDomainsOrRanges = next.getDirectDomain();
				nextDomainsOrRanges = getDomains (next);
			} else if (next.getValueType() == ValueType.INSTANCE) {
				nextDomainsOrRanges = next.getAllowedClses();
			} else {
				// *** This works in the case of the FMA, but not in general
				nextDomainsOrRanges = next.getAllowedParents();
			}
			
			Collection owlDomainsOrRanges = FMAtoOWL.getOWLClasses(nextDomainsOrRanges);
			if (owlDomainsOrRanges != null && !owlDomainsOrRanges.isEmpty()) {
				if (owlDomainsOrRanges.size() == 1) {
					OWLNamedClass owlDomainOrRange = (OWLNamedClass)CollectionUtilities.getSoleItem (owlDomainsOrRanges);
					if (domainOrRange == DOMAIN)
						nextProperty.setDomain(owlDomainOrRange);
					else
						nextProperty.setRange(owlDomainOrRange);
				} else {
					if (domainOrRange == DOMAIN)
						nextProperty.setDomain(_owlModel.createOWLUnionClass(owlDomainsOrRanges));
					else
						nextProperty.setUnionRangeClasses(owlDomainsOrRanges);
				}
			}
		}
		
	}
	
	//**** This is a workaround; slot.getDirectDomain sholud work instead
	//but there is a Protege bug, and for many slots it returns an incomplete list
	// (e.g. regional part doesn't have Anatomical structure in its domain)
	private Collection getDomains (Slot slot) {
		Collection result = new ArrayList ();
		Collection references = slot.getReferences();
		
		
		Iterator r = references.iterator();
		while (r.hasNext()) {
			Reference next = (Reference)r.next();
			if (!next.getSlot().equals(_directTemplateSlot)) continue;
			result.add(next.getFrame());
		}
		
		return result;
	}


	private void firstPass (String [] roots) {
		
		processRootClses (roots);
	    
		int counter = 0;
		
		while (!_classesInQueue.isEmpty()) {
	        	Cls next = (Cls)_classesInQueue.firstElement();
	        	_classesInQueue.remove(0);
	        	
	     	OWLNamedClass nextOwlClass = FMAtoOWL.getOwlClass (next);
	        	if (nextOwlClass != null) continue;
	        	
	        	if (counter % 1000 == 0)
	        		Log.getLogger().info("First pass (" + _fromFullString + "): " + counter + ":" + next);
	
	        	nextOwlClass = createOWLCopy (next);
	        	
	        	
	         	Collection nextSubs = next.getDirectSubclasses();
	        	_classesInQueue.addAll(nextSubs);
	        	counter++;
	        }
		
	 }
	
	private void secondPass () {
		Iterator i = _sourceKB.getClses().iterator();
		int counter = 0;
		while (i.hasNext()) {
			Object nextObject = i.next();
			if (!(nextObject instanceof Cls)) {
				Log.getLogger().info("Not a class: " + nextObject);
				continue;
			}
			Cls next = (Cls)nextObject;
			if (next.isSystem()) continue;

			if (counter % 1000 == 0)
        			Log.getLogger().info("Second pass (" + _fromFullString + "): " + counter + ":" + next);

			
			OWLNamedClass nextOWLClass = FMAtoOWL.getOwlClass(next);
			if (nextOWLClass == null) continue;
			
			copyClassSuperclasses (next, nextOWLClass);
			addRestrictionsForOverrides (next, nextOWLClass);
			copyObjectAnnatationProperties (next, nextOWLClass);
			counter++;
		}
		
		
	}

	private void addRestrictionsForOverrides(Cls cls, OWLNamedClass result) {
		Collection directSlots = cls.getTemplateSlots();
		if (directSlots == null || directSlots.isEmpty()) return;
		Iterator s = directSlots.iterator();
		while (s.hasNext()) {
			Slot nextSlot = (Slot)s.next();
			
			OWLProperty nextProperty = FMAtoOWL.getOWLProperty(nextSlot);
			if (nextProperty == null) continue;
			
			Collection overriddenFacets = cls.getDirectlyOverriddenTemplateFacets(nextSlot);
			if (overriddenFacets == null || overriddenFacets.isEmpty()) continue;
			
			Iterator f = overriddenFacets.iterator();
			while (f.hasNext()) {
				Facet nextFacet = (Facet)f.next();
				if (nextFacet.getFrameID().equals(Model.FacetID.VALUES)) {
					Collection values = cls.getDirectTemplateFacetValues(nextSlot, nextFacet);
					if (values == null || values.isEmpty()) continue;
					Iterator v = values.iterator();
					while (v.hasNext()) {
						Object nextValue = v.next();
						if (nextValue instanceof Cls)
							continue;
		//					nextValue = FMAtoOWL.getOwlClass((Cls)nextValue);
						result.addSuperclass(_owlModel.createOWLHasValue(nextProperty, nextValue));
					}
				} else if (nextFacet.getFrameID().equals(Model.FacetID.VALUE_TYPE)) {
					Collection values = cls.getTemplateSlotAllowedClses(nextSlot);
					Collection owlValues = FMAtoOWL.getOWLClasses (values);
					if (owlValues == null  || owlValues.isEmpty()) continue;
					if (owlValues.size() == 1)
						result.addSuperclass(_owlModel.createOWLAllValuesFrom(nextProperty, 
								(OWLNamedClass)CollectionUtilities.getSoleItem(owlValues)));
					else	
						result.addSuperclass(_owlModel.createOWLAllValuesFrom(nextProperty, 
								_owlModel.createOWLUnionClass(owlValues)));
				} else if (nextFacet.getFrameID().equals(Model.FacetID.MINIMUM_CARDINALITY)) {
					int min = cls.getTemplateSlotMinimumCardinality(nextSlot);
					result.addSuperclass(_owlModel.createOWLMinCardinality(nextProperty, min));
				} else if (nextFacet.getFrameID().equals(Model.FacetID.MAXIMUM_CARDINALITY)) {
					int min = cls.getTemplateSlotMaximumCardinality(nextSlot);
					result.addSuperclass(_owlModel.createOWLMaxCardinality(nextProperty, min));
				} else {
					Log.getLogger().warning("Not processing facet: " + nextFacet + " for class "+ cls + " for slot " + nextSlot);
				}
					
			}
		}
		
		
	}

	private OWLNamedClass createOWLCopy (Cls cls) {
		OWLNamedClass result = _owlModel.createOWLNamedClass (FMAtoOWL.getValidOWLName (cls.getName()));
//		FMAtoOWL.addOWLClassName(cls, result);
		
//		String label = null;
//		if (_rdfsLabelSlot != null) {
//			label = (String)cls.getOwnSlotValue(_rdfsLabelSlot);
//		}
//		result.addLabel((label == null) ? cls.getName() : label, Constants.LANGUAGE_LABEL);     
		

		result.addLabel(cls.getBrowserText(), Constants.LANGUAGE_LABEL);
		copyDatatypeAnnotationProperties (cls, result);
		
//		copyClassSuperclasses (cls, result);
	
		return result;
		
	}

	private void copyDatatypeAnnotationProperties(Cls cls, OWLNamedClass result) {
		Collection clsDocumentation = cls.getDocumentation();
		if (clsDocumentation != null)
			result.setComment((String)CollectionUtilities.getFirstItem (clsDocumentation));
	
		for (int i = 0; i < Constants.ANNOTATION_DATATYPE_PROPERTIES.length; i++) {
			Slot slot = _sourceKB.getSlot(Constants.ANNOTATION_DATATYPE_PROPERTIES[i]);
			OWLProperty property = FMAtoOWL.getOWLProperty(slot);
			Collection values = cls.getOwnSlotValues(slot);
			if (values != null) {
				Iterator v = values.iterator();
				while (v.hasNext()) {
					String next = (String)v.next();
					result.addPropertyValue(property, new DefaultRDFSLiteral (_owlModel, next));
				}
			}
		}
		
	}

	private void copyClassSuperclasses (Cls cls, OWLNamedClass result) {
	   	Collection supers = cls.getDirectSuperclasses();
	   	Iterator i = supers.iterator();
	   	while (i.hasNext()) {
	   		Cls nextSuper = (Cls)i.next();
	   		if (nextSuper == _sourceKB.getRootCls()) continue;
	   		
	    		OWLNamedClass nextOwlSuper = FMAtoOWL.getOwlClass (nextSuper);
	   		if (nextOwlSuper != null) {
	   			result.addSuperclass(nextOwlSuper);
	   			result.removeSuperclass(_owlModel.getOWLThingClass());
	   		}
	   	}	
	}

	// assuming that all values of these properties are simple instances and slot values  there are strings or booleans
	// (as is the case for instances of Concept name
	private void copyObjectAnnatationProperties(Cls cls, OWLNamedClass result) {
		for (int i = 0; i < Constants.ANNOTATION_OBJECT_PROPERTIES.length; i++) {
			Slot slot = _sourceKB.getSlot(Constants.ANNOTATION_OBJECT_PROPERTIES[i]);
			OWLProperty property = FMAtoOWL.getOWLProperty(slot);
			Collection values = cls.getOwnSlotValues(slot);
			if (values != null) {
				Iterator v = values.iterator();
				while (v.hasNext()) {
					Instance next = (Instance)v.next();
					Instance owlInstance = Util.copyInstance (_owlModel, next);
					if (owlInstance != null)
						result.addPropertyValue(property, owlInstance);
				}
			}
		}		
	}

	private void processRootClses (String [] roots) {
		for (int i = 0; i < roots.length; i++) {
			Cls next = _sourceKB.getCls(roots[i]);
			createOWLCopy (next);
			_classesInQueue.addAll(next.getDirectSubclasses());
		}
	}
	

}
